﻿.include "FPP.inc"
	
	jp Main

Test
	.db "(-8>>2)+ceil(pi())+200E-2**sqrt(abs((~(2&4)>>>(30^sin(rad(90))))-(10>?1)))",0

Accumulator  = OP1+0 ; 5 bytes for the floating-point accumulator.
Temp16       = OP1+5 ; 2 bytes to store a temporary 16-bit value.
EntrySP      = OP1+7 ; 2 bytes to store SP on entry.
FPPPage      = OP2+0 ; 1 byte to store the ROM page of the FPP routine.
CurrentPage  = OP2+1 ; 1 byte to store the current ROM page.
OutputFormat = OP2+2 ; 2 byte string format.

ErrorStrings
.db ERR_BAD_OPERATION,  "Bad operation", 0
.db ERR_MISSING_LPAREN, "Missing (", 0
.db ERR_SYNTAX,         "Syntax", 0
.db ERR_DIV_BY_ZERO,    "Division by zero", 0
.db ERR_TOO_BIG,        "Too big", 0
.db ERR_NEGATIVE_ROOT,  "-ve root", 0
.db ERR_LOG_RANGE,      "Log range", 0
.db ERR_ACCURACY_LOST,  "Accuracy lost", 0
.db ERR_EXP_RANGE,      "Exp range", 0
.db ERR_MISSING_RPAREN, "Missing )", 0
.db 0,                  "Unknown error", 0

ErrorBBCBasicMissing
.db "BBC BASIC not found", 0

Main

	; Initialise.

	.bcall _ClrLcdFull
	ld hl,0
	ld (curRow),a
	
	call Initialise
	jr nc,+
	ld hl,ErrorBBCBasicMissing
	.bcall _PutS
	ret
+

	; Evaluate.
	
	ld ix,Test
	call Evaluate
	ei
	jr nc,NoError
	
	; There was an error in the input.
	ld c,a
	ld hl,ErrorStrings
SearchErrorString
	ld a,(hl)
	inc hl
	cp c
	jr z,FoundErrorString
	
-	ld a,(hl)
	inc hl
	or a
	jr nz,-
	
	ld a,(hl)
	or a
	jr nz,SearchErrorString
	inc hl
	
FoundErrorString

	.bcall _PutS
	.bcall _NewLine
	ret

NoError
	
	; There wasn't an error; display the answer.
	ld hl,$0A00
	ld (OutputFormat),hl
	ld de,OP3
	call ToString
	ei
	
	ld hl,OP3
	.bcall _PutS
	.bcall _NewLine
	ret

; ========================================================================== ;
; Initialises the evaluator. Sets C on error.                                ;
; ========================================================================== ;
Initialise
	ld hl,AppName
	rst rMov9ToOP1
	.bcall _FindApp
	ret c
	dec a
	ld (FPPPage),a
	or a
	ret

AppName
.db AppObj, "BBCBasic"

; ========================================================================== ;
; Evaluate an expression pointed to by IX. Sets C on error with error code   ;
; in A. Disabled interrupts.                                                 ;
; ========================================================================== ;
Evaluate
	di
	in a,(memPageAPort)
	ld (CurrentPage),a
	ld a,(FPPPage)
	out (memPageAPort),a
	ld (EntrySP),sp
	call Evaluate.Top
	ld a,(CurrentPage)
	out (memPageAPort),a
	or a
	ret

; ========================================================================== ;
; Level: Top level                                                           ;
; ========================================================================== ;	
Evaluate.Top

; ========================================================================== ;
; Level: LogicalOr                                                           ;
; ========================================================================== ;
Evaluate.LogicalOr
	call Evaluate.LogicalXor
Evaluate.LogicalOr.Loop

	call ReadCharacter
	ret z
	
	cp '|'
	ret nz

; -------------------------------------------------------------------------- ;
; Operator: Or (|)                                                           ;
; -------------------------------------------------------------------------- ;
Evaluate.LogicalOr.Or
	inc ix
	call PushAccumulator
	call Evaluate.LogicalXor
	call PopOP1
	call LoadOP2
	ld a,FPP_OR
	call FPP_PushIX
	jp c,Error
	call StoreOP1
	jr Evaluate.LogicalOr.Loop

; ========================================================================== ;
; Level: LogicalXor                                                          ;
; ========================================================================== ;
Evaluate.LogicalXor
	call Evaluate.LogicalAnd
Evaluate.LogicalXor.Loop

	call ReadCharacter
	ret z
	
	cp '^'
	ret nz

; -------------------------------------------------------------------------- ;
; Operator: Xor (^)                                                          ;
; -------------------------------------------------------------------------- ;
Evaluate.LogicalXor.Xor
	inc ix
	call PushAccumulator
	call Evaluate.LogicalAnd
	call PopOP1
	call LoadOP2
	ld a,FPP_EOR
	call FPP_PushIX
	jp c,Error
	call StoreOP1
	jr Evaluate.LogicalXor.Loop

; ========================================================================== ;
; Level: LogicalAnd                                                          ;
; ========================================================================== ;
Evaluate.LogicalAnd
	call Evaluate.Equality
Evaluate.LogicalAnd.Loop

	call ReadCharacter
	ret z
	
	cp '&'
	ret nz

; -------------------------------------------------------------------------- ;
; Operator: And (&)                                                          ;
; -------------------------------------------------------------------------- ;
Evaluate.LogicalAnd.And
	inc ix
	call PushAccumulator
	call Evaluate.Equality
	call PopOP1
	call LoadOP2
	ld a,FPP_AND
	call FPP_PushIX
	jp c,Error
	call StoreOP1
	jr Evaluate.LogicalAnd.Loop	

; ========================================================================== ;
; Level: Equality (== !=)                                                    ;
; ========================================================================== ;
Evaluate.Equality
	call Evaluate.Relational
Evaluate.Equality.Loop

	call ReadCharacter
	ret z
	
	cp '='
	jr z,Evaluate.Equality.Equality
	cp '!'
	jr z,Evaluate.Equality.Inequality
	
	ret

; -------------------------------------------------------------------------- ;
; Operator: Equality (==)                                                    ;
; -------------------------------------------------------------------------- ;
Evaluate.Equality.Equality
	ld a,(ix+1)
	cp '='
	ret nz
	
	inc ix
	inc ix
	call PushAccumulator
	call Evaluate.Relational
	call PopOP1
	call LoadOP2
	ld a,FPP_EQ
	call FPP_PushIX
	jp c,Error
	call StoreOP1
	jr Evaluate.Equality.Loop
	
; -------------------------------------------------------------------------- ;
; Operator: Inequality (!=)                                                  ;
; -------------------------------------------------------------------------- ;
Evaluate.Equality.Inequality
	ld a,(ix+1)
	cp '='
	ret nz
	
	inc ix
	inc ix
	call PushAccumulator
	call Evaluate.Relational
	call PopOP1
	call LoadOP2
	ld a,FPP_INEQ
	call FPP_PushIX
	jp c,Error
	call StoreOP1	
	jr Evaluate.Equality.Loop

; ========================================================================== ;
; Level: Relational (< > <= >= <? ?>)                                        ;
; ========================================================================== ;
Evaluate.Relational
	call Evaluate.Shift
Evaluate.Relational.Loop

	call ReadCharacter
	ret z
	
	cp '<'
	jr z,Evaluate.Relational.LessThan
	cp '>'
	jr z,Evaluate.Relational.GreaterThan
	ret

; -------------------------------------------------------------------------- ;
; Operator: Less than (<)                                                    ;
; -------------------------------------------------------------------------- ;
Evaluate.Relational.LessThan
	inc ix
	ld a,(ix)
	cp '='
	jr z,Evaluate.Relational.LessThanEqual
	cp '?'
	jr z,Evaluate.Relational.Minimum
	
	call PushAccumulator
	call Evaluate.Shift
	call PopOP1
	call LoadOP2
	ld a,FPP_LT
	call FPP_PushIX
	jp c,Error
	call StoreOP1
	jr Evaluate.Relational.Loop

; -------------------------------------------------------------------------- ;
; Operator: Minimum (<?)                                                     ;
; -------------------------------------------------------------------------- ;
Evaluate.Relational.Minimum
	inc ix
	call PushAccumulator
	call Evaluate.Shift
	call PopOP2
	call PushOP2
	call LoadOP1

	ld a,FPP_COMPARE
	call FPP_PushIX
	jp c,Error

	cp $40
	call PopOP1
	jp nz,Evaluate.Relational.Loop
	call StoreOP1
	jp Evaluate.Relational.Loop


; -------------------------------------------------------------------------- ;
; Operator: Less than or equal to (<=)                                       ;
; -------------------------------------------------------------------------- ;
Evaluate.Relational.LessThanEqual
	inc ix

	call PushAccumulator
	call Evaluate.Shift
	call PopOP1
	call LoadOP2
	ld a,FPP_LTEQ
	call FPP_PushIX
	jp c,Error
	call StoreOP1
	jr Evaluate.Relational.Loop
	jr Evaluate.Relational.Loop

; -------------------------------------------------------------------------- ;
; Operator: Greater than (>)                                                 ;
; -------------------------------------------------------------------------- ;
Evaluate.Relational.GreaterThan
	inc ix
	ld a,(ix)
	cp '='
	jr z,Evaluate.Relational.GreaterThanEqual
	cp '?'
	jr z,Evaluate.Relational.Maximum
	
	call PushAccumulator
	call Evaluate.Shift
	call PopOP1
	call PushOP1
	call LoadOP2
	ld a,FPP_GT
	call FPP_PushIX
	jp c,Error
	call StoreOP1
	jp Evaluate.Relational.Loop

; -------------------------------------------------------------------------- ;
; Operator: Greater than or equal to (>=)                                    ;
; -------------------------------------------------------------------------- ;
Evaluate.Relational.GreaterThanEqual
	inc ix
	
	call PushAccumulator
	call Evaluate.Shift
	call PopOP1
	call LoadOP2
	ld a,FPP_GTEQ
	call FPP_PushIX
	jp c,Error
	call StoreOP1
	jp Evaluate.Relational.Loop

; -------------------------------------------------------------------------- ;
; Operator: Maximum (>?)                                                     ;
; -------------------------------------------------------------------------- ;
Evaluate.Relational.Maximum
	inc ix
	call PushAccumulator
	call Evaluate.Shift
	call PopOP2
	call PushOP2
	call LoadOP1

	ld a,FPP_COMPARE
	call FPP_PushIX
	jp c,Error

	cp $C0
	call PopOP1
	jp nz,Evaluate.Relational.Loop
	call StoreOP1
	jp Evaluate.Relational.Loop

; ========================================================================== ;
; Level: Shift (<< >> >>>)                                                   ;
; ========================================================================== ;
Evaluate.Shift
	call Evaluate.Additive
Evaluate.Shift.Loop

	call ReadCharacter
	ret z
	
	cp (ix+1)
	ret nz
	
	cp '<'
	jr z,Evaluate.Shift.ShiftLeft
	cp '>'
	jr z,Evaluate.Shift.ShiftRight
	ret

; -------------------------------------------------------------------------- ;
; Operator: Shift left (<<)                                                  ;
; -------------------------------------------------------------------------- ;
Evaluate.Shift.ShiftLeft
	inc ix
	inc ix
	
	call Evaluate.Shift.GetOperands

	jr z,+
	ld b,a
-	sla l
	rl h
	rl e
	rl d
	djnz -
+	ld (Accumulator+0),hl
	ld (Accumulator+2),de	
	jr Evaluate.Shift.Loop

; -------------------------------------------------------------------------- ;
; Operator: Shift right (>>)                                                 ;
; -------------------------------------------------------------------------- ;
Evaluate.Shift.ShiftRight
	inc ix
	inc ix

	ld a,(ix)
	cp '>'
	jr z,Evaluate.Shift.ShiftRightUnsigned
	
	call Evaluate.Shift.GetOperands

	jr z,+
	ld b,a
-	sra d
	rr e
	rr h
	rr l
	djnz -
+	ld (Accumulator+0),hl
	ld (Accumulator+2),de	
	jp Evaluate.Shift.Loop

; -------------------------------------------------------------------------- ;
; Operator: Shift right unsigned (>>>)                                       ;
; -------------------------------------------------------------------------- ;
Evaluate.Shift.ShiftRightUnsigned
	inc ix
	
	call Evaluate.Shift.GetOperands

	jr z,+
	ld b,a
-	srl d
	rr e
	rr h
	rr l
	djnz -
+	ld (Accumulator+0),hl
	ld (Accumulator+2),de	
	jr Evaluate.Shift.Loop

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  ;
; Get shift operands: DEHL is the integer to shift, A is number of places to ;
; shift it and Z is set if A = 0.                                            ;
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -  ;
Evaluate.Shift.GetOperands
	call LoadOP1
	ld a,FPP_FIX
	call FPP_PushIX
	jp c,Error
	call PushOP1
	
	call Evaluate.Additive
	
	call LoadOP1
	ld a,FPP_FIX
	call FPP_PushIX
	jp c,Error
	
	call PopOP2
	
	exx
	ld a,l
	push de
	exx
	pop hl
	
	and 31
	ret

; ========================================================================== ;
; Level: Additive (+ -)                                                      ;
; ========================================================================== ;
Evaluate.Additive
	call Evaluate.Multiplicative
Evaluate.Additive.Loop
	
	call ReadCharacter
	ret z
	
	cp '+'
	jp z,Evaluate.Additive.Add
	cp '-'
	jp z,Evaluate.Additive.Subtract
	ret

; -------------------------------------------------------------------------- ;
; Operator: Add (+)                                                          ;
; -------------------------------------------------------------------------- ;
Evaluate.Additive.Add
	inc ix
	call PushAccumulator
	call Evaluate.Multiplicative	
	call PopOP1
	call LoadOP2
	ld a,FPP_ADD
	call FPP_PushIX
	jp c,Error
	call StoreOP1	
	jp Evaluate.Additive.Loop

; -------------------------------------------------------------------------- ;
; Operator: Subtract (-)                                                     ;
; -------------------------------------------------------------------------- ;
Evaluate.Additive.Subtract
	inc ix
	call PushAccumulator	
	call Evaluate.Multiplicative	
	call PopOP1
	call LoadOP2	
	ld a,FPP_SUB
	call FPP_PushIX
	jp c,Error
	call StoreOP1	
	jp Evaluate.Additive.Loop

; ========================================================================== ;
; Level: Multiplicative (* / \ %)                                            ;
; ========================================================================== ;
Evaluate.Multiplicative
	call Evaluate.Power
Evaluate.Multiplicative.Loop
	
	call ReadCharacter
	ret z
	
	cp '*'
	jr z,Evaluate.Multiplicative.Multiply
	cp '/'
	jr z,Evaluate.Multiplicative.FDivide
	cp '\\'
	jr z,Evaluate.Multiplicative.IDivide
	cp '%'
	jr z,Evaluate.Multiplicative.Modulo
	ret	

; -------------------------------------------------------------------------- ;
; Operator: Multiply (*)                                                     ;
; -------------------------------------------------------------------------- ;
Evaluate.Multiplicative.Multiply
	inc ix
	call PushAccumulator
	call Evaluate.Power	
	call PopOP1
	call LoadOP2
	ld a,FPP_MUL
	call FPP_PushIX
	jp c,Error
	call StoreOP1	
	jp Evaluate.Multiplicative.Loop

; -------------------------------------------------------------------------- ;
; Operator: Floating point divide (/)                                        ;
; -------------------------------------------------------------------------- ;
Evaluate.Multiplicative.FDivide
	inc ix
	call PushAccumulator
	call Evaluate.Power	
	call PopOP1
	call LoadOP2
	ld a,FPP_FDIV
	call FPP_PushIX
	jp c,Error
	call StoreOP1	
	jp Evaluate.Multiplicative.Loop

; -------------------------------------------------------------------------- ;
; Operator: Integer divide (\)                                               ;
; -------------------------------------------------------------------------- ;
Evaluate.Multiplicative.IDivide
	inc ix
	call PushAccumulator
	call Evaluate.Power	
	call PopOP1
	call LoadOP2
	ld a,FPP_IDIV
	call FPP_PushIX
	jp c,Error
	call StoreOP1	
	jp Evaluate.Multiplicative.Loop

; -------------------------------------------------------------------------- ;
; Operator: Modulo (%)                                                       ;
; -------------------------------------------------------------------------- ;
Evaluate.Multiplicative.Modulo
	inc ix
	call PushAccumulator
	call Evaluate.Power	
	call PopOP1
	call LoadOP2
	ld a,FPP_MOD
	call FPP_PushIX
	jp c,Error
	call StoreOP1	
	jp Evaluate.Multiplicative.Loop

; ========================================================================== ;
; Level: Power (**)                                                          ;
; ========================================================================== ;
Evaluate.Power
	call Evaluate.Literal
Evaluate.Power.Loop

	call ReadCharacter
	ret z
	cp '*'
	ret nz
	ld a,(ix+1)
	cp '*'
	ret nz

; -------------------------------------------------------------------------- ;
; Operator: Power (**)                                                       ;
; -------------------------------------------------------------------------- ;
Evaluate.Power.Power
	inc ix
	inc ix
	call PushAccumulator
	call Evaluate.Literal	
	call PopOP1
	call LoadOP2
	ld a,FPP_POW
	call FPP_PushIX
	jp c,Error
	call StoreOP1	
	jr Evaluate.Power.Loop	

; ========================================================================== ;
; Level: Literals (also handles unary +/- and parens).                       ;
; ========================================================================== ;
Evaluate.Literal
	call ReadCharacter
	jr nz,+
	ld a,ERR_SYNTAX
	jp Error
+	cp '('
	jr nz,Evaluate.Literal.NotParen
	inc ix
	call Evaluate.Top ; Recursively evaluate expressions.
	call ReadCharacter
	cp ')'
	ld a,ERR_MISSING_RPAREN
	jp nz,Error
	inc ix
	ret
Evaluate.Literal.NotParen

	cp '-'
	jr nz,Evaluate.Literal.NotUnaryNegate
	
	inc ix
	call Evaluate.Literal
	call LoadOP2
	ld a,FPP_ZERO
	call FPP_PushIX
	jp c,Error
	ld a,FPP_SUB
	call FPP_PushIX
	jp c,Error
	call StoreOP1
	ret
	
Evaluate.Literal.NotUnaryNegate

	cp '+'
	jr nz,Evaluate.Literal.NotUnaryAdd
	inc ix
	jr Evaluate.Literal
Evaluate.Literal.NotUnaryAdd

	cp '~'
	jr nz,Evaluate.Literal.NotComplement
	
	inc ix
	call Evaluate.Literal
	call LoadOP1
	ld a,FPP_NOT
	call FPP_PushIX
	jp c,Error
	call StoreOP1
	ret
	
Evaluate.Literal.NotComplement

	cp '!'
	jr nz,Evaluate.Literal.NotNot

	inc ix
	call Evaluate.Literal

	ld a,FPP_ZERO
	call FPP_PushIX
	jp c,Error
	call LoadOP2
	
	ld a,FPP_EQ
	call FPP_PushIX
	jp c,Error
	call StoreOP1
	ret

Evaluate.Literal.NotNot

	; Search through function names.

	ld hl,Evaluate.Function.Names
Evaluate.Literal.SearchFunction	

	ld b,(hl)
	inc hl
	
	push ix
	
-	ld a,(hl)
	inc hl
	cp (ix)
	inc ix
	jr nz,+
	djnz -
+	push af
	ld c,b
	ld b,0
	add hl,bc
	pop af

	jr z,Evaluate.Function
	pop ix
	inc hl
	inc hl
	ld a,(hl)
	or a
	jr z,Evaluate.Literal.IsNumeric
	jr Evaluate.Literal.SearchFunction

Evaluate.Function
	pop af ; Ignore IX
	ld a,(ix)
	cp '('
	jr z,+
	ld a,ERR_MISSING_LPAREN
	jp Error
+	inc ix
	ld a,(hl)
	inc hl
	ld e,(hl)
	inc hl
	ld d,(hl)
	ex de,hl
	jp (hl)

Evaluate.Literal.IsNumeric
	push ix
	ld a,FPP_VAL
	call FPP
	call StoreOP1
	pop hl
	push ix
	pop de
	or a
	sbc hl,de
	ret nz
	ld a,ERR_SYNTAX
	jp Error

; ========================================================================== ;
; Returns the ceiling of a number.                                           ;
; ========================================================================== ;
Evaluate.Function.Ceil
	call Evaluate.Top
	call ReadRightParen
	ld a,FPP_ZERO
	call FPP_PushIX
	call LoadOP2
	ld a,FPP_SUB
	call FPP_PushIX
	jp c,Error
	ld a,FPP_INT
	call FPP_PushIX
	jp c,Error
	ex de,hl
	exx
	ex de,hl
	exx
	ld b,c
	ld a,FPP_ZERO
	call FPP_PushIX
	ld a,FPP_SUB
	call FPP_PushIX
	jp c,Error
	jp StoreOP1

; ========================================================================== ;
; Returns π                                                                  ;
; ========================================================================== ;
Evaluate.Function.Pi
	call FPP_PushIX
	call StoreOP1
	jp ReadRightParen	

; ========================================================================== ;
; Execute a function that takes a single numeric argument.                   ;
; ========================================================================== ;
Evaluate.Function.Execute
	push af
	call Evaluate.Top
	call ReadRightParen
	call LoadOP1
	pop af
	call FPP_PushIX
	jp c,Error
	jp StoreOP1

Evaluate.Function.Names
.db 3, "abs",   FPP_ABS \ .dw Evaluate.Function.Execute
.db 4, "acos",  FPP_ACS \ .dw Evaluate.Function.Execute
.db 4, "asin",  FPP_ASN \ .dw Evaluate.Function.Execute
.db 4, "atan",  FPP_ATN \ .dw Evaluate.Function.Execute
.db 4, "ceil",  0       \ .dw Evaluate.Function.Ceil
.db 3, "cos",   FPP_COS \ .dw Evaluate.Function.Execute
.db 3, "deg",   FPP_DEG \ .dw Evaluate.Function.Execute
.db 3, "exp",   FPP_EXP \ .dw Evaluate.Function.Execute
.db 5, "floor", FPP_INT \ .dw Evaluate.Function.Execute
.db 3, "int",   FPP_FIX \ .dw Evaluate.Function.Execute
.db 2, "ln",    FPP_LN  \ .dw Evaluate.Function.Execute
.db 3, "log",   FPP_LOG \ .dw Evaluate.Function.Execute
.db 2, "pi",    FPP_PI  \ .dw Evaluate.Function.Pi
.db 3, "rad",   FPP_RAD \ .dw Evaluate.Function.Execute
.db 3, "sgn",   FPP_SGN \ .dw Evaluate.Function.Execute
.db 3, "sin",   FPP_SIN \ .dw Evaluate.Function.Execute
.db 4, "sqrt",  FPP_SQR \ .dw Evaluate.Function.Execute
.db 3, "tan",   FPP_TAN \ .dw Evaluate.Function.Execute
.db 0

; ========================================================================== ;
; Reads a character from the input string, skipping whitespace. Returns Z if ;
; the end of the input has been reached.                                     ;
; ========================================================================== ;
ReadCharacter
	ld a,(ix)
	cp ' '
	jr z,+
	or a
	ret
+	inc ix
	jr ReadCharacter

; ========================================================================== ;
; Reads a right parenthesis character from the input string.                 ;
; ========================================================================== ;
ReadRightParen
	call ReadCharacter
	cp ')'
	jr nz,+
	inc ix
	ret
+	ld a,ERR_MISSING_RPAREN
	jp Error

; ========================================================================== ;
; Store operand 1 to the accumulator                                         ;
; ========================================================================== ;
StoreOP1
	ld (Accumulator+2),hl
	exx
	ld (Accumulator+0),hl
	exx
	ld a,c
	ld (Accumulator+4),a
	ret

; ========================================================================== ;
; Load operand 1 from the accumulator                                        ;
; ========================================================================== ;
LoadOP1
	ld hl,(Accumulator+2)
	exx
	ld hl,(Accumulator+0)
	exx
	ld a,(Accumulator+4)
	ld c,a
	ret

; ========================================================================== ;
; Store operand 2 to the accumulator                                         ;
; ========================================================================== ;
StoreOP2
	ld (Accumulator+2),de
	exx
	ld (Accumulator+0),de
	exx
	ld a,b
	ld (Accumulator+4),a
	ret

; ========================================================================== ;
; Load operand 2 from the accumulator                                        ;
; ========================================================================== ;
LoadOP2
	ld de,(Accumulator+2)
	exx
	ld de,(Accumulator+0)
	exx
	ld a,(Accumulator+4)
	ld b,a
	ret

; ========================================================================== ;
; Push operand 1 onto the stack                                              ;
; ========================================================================== ;
PushOP1
	ld (Temp16),de
	pop de
	exx
	push hl
	exx
	push hl
	push bc
	push de
	ld de,(Temp16)
	ret

; ========================================================================== ;
; Pop operand 1 off the stack                                                ;
; ========================================================================== ;
PopOP1
	ld (Temp16),de
	pop de
	ld a,b
	pop bc
	ld b,a
	pop hl
	exx
	pop hl
	exx
	push de
	ld de,(Temp16)
	ret

; ========================================================================== ;
; Push operand 2 onto the stack                                              ;
; ========================================================================== ;
PushOP2
	ld (Temp16),hl
	pop hl
	exx
	push de
	exx
	push de
	push bc
	push hl
	ld hl,(Temp16)
	ret

; ========================================================================== ;
; Pop operand 2 off the stack                                                ;
; ========================================================================== ;
PopOP2
	ld (Temp16),hl
	pop hl
	ld a,c
	pop bc
	ld b,c
	ld c,a
	pop de
	exx
	pop de
	exx
	push hl
	ld hl,(Temp16)
	ret

; ========================================================================== ;
; Push accumulator onto the stack                                            ;
; ========================================================================== ;
PushAccumulator
	pop hl
	ld de,(Accumulator+0)
	push de
	ld de,(Accumulator+2)
	push de
	ld de,(Accumulator+4)
	push de	
	jp (hl)

; ========================================================================== ;
; Pop accumulator off the stack                                              ;
; ========================================================================== ;
PopAccumulator
	pop hl
	pop de
	ld a,e
	ld (Accumulator+4),a
	pop de
	ld (Accumulator+2),de
	pop de
	ld (Accumulator+0),de
	jp (hl)

; ========================================================================== ;
; Report an error                                                            ;
; ========================================================================== ;
Error
	push af
	ld a,(CurrentPage)
	out (6),a
	pop af
	scf
	ld sp,(EntrySP)
	ret

; ========================================================================== ;
; Converts the accumulator to a string. Disables interrupts.                 ;
; ========================================================================== ;
ToString
	di
	ld a,(FPPPage)
	out (memPageAPort),a
	call LoadOP1
	ld ix,OutputFormat-1
	ld a,FPP_STR$
	call FPP_PushIX
	xor a
	ld (de),a
	ld a,(CurrentPage)
	out (memPageAPort),a
	ret

; ========================================================================== ;
; Calls FPP but preserves IX.                                                ;
; ========================================================================== ;
FPP_PushIX
	push ix
	call FPP
	pop ix
	ret